from multiprocess import Pool
import dill

def timeout(sec=1):
    def wrapper(f):
        def inner(*args):
            pool = Pool(processes=1)
            res = pool.apply_async(f,args)
            try:
                v = res.get(timeout=sec)
            except Exception as inst:
                print(inst)
                v = None
            finally:
                pool.terminate()
                return v
        return inner
    return wrapper

def loader(pkl,args):
    f = dill.loads(pkl)
    return f(*args)

def timeout_dumb(sec=1):
    def wrapper(pkl):
        def inner(*args):
            pool = Pool(processes=1)
            res = pool.apply_async(loader,(pkl,args))
            try:
                v = res.get(timeout=sec)
            except Exception as inst:
                # print(inst)
                v = None
            finally:
                pool.terminate()
                return v
        return inner
    return wrapper



# AUTO_PATCH=True python test_backoff.py
import collections
import math
from context import Meta

meta = Meta(globals=globals())

# @meta("get most frequent character from string")
# def get_most_frequent_character(string):
#     if string == "": return None
#     return collections.Counter(string).most_common(1)[0]
#
# get_most_frequent_character("ethan fast")
#
# get_most_frequent_character.get_dynamic_type_sig()

# @meta("extract all digits from a string")
# def extract_digits(n):
# 	return [int(x) for x in str(n) if x in "0123456789"]
#
# extract_digits("123 45")
#
# extract_digits.get_dynamic_type_sig()

# @meta("calculate log of each element in array")
# def log_array(a):
# 	return [math.log(x) for x in a]
#
# @meta("calculate log of each element in array, with error as zero")
# def log_array_safe(a):
# 	return [math.log(x) if x > 0 else 0 for x in a]
#
# log_array([2,3,4])
# log_array_safe([6,8,12])
#
# log_array_safe.get_dynamic_type_sig()
#
# print(log_array.find_duplicates())

log_array = meta.load("http://www.meta-lang.org/snippets/570a91def842a50003be6280")

log_array([1,2,0,4])

get_most_frequent_character = meta.load("http://www.meta-lang.org/snippets/56ee4681c0cb8f7470fbb317")

get_most_frequent_character("")

convert_string_to_int = meta.load("http://www.meta-lang.org/snippets/56ee44fcc0cb8f7470fbaf23")

convert_string_to_int("abc") # patch with safe string to int

convert_string_to_float = meta.load("http://www.meta-lang.org/snippets/56ee44fcc0cb8f7470fbaf21")

# here's a bug
print(convert_string_to_float("ethan")) # patch with safe string to float



from context import Meta
from collections import Counter
import numpy

meta = Meta(overhead=True)

@meta("add two numbers")
def add_numbers(x,y):
    return x + y

@meta("count words in a text document", imports=[{"from":"collections", "import":"Counter"}])
def count_words(doc):
    words = [w.lower() for w in doc.split()]
    return Counter(words)

@meta("compute difference in average between lists")
def average_list_difference(l1,l2):
    l1_avg, l2_avg = [numpy.average(x) for x in [l1,l2]]
    return l1_avg - l2_avg

average_list_difference([1,2,3],[3,4,5])

print(average_list_difference.__meta_imports__)

count_words("this is a test document")

print("Meta id: {}".format(add_numbers.__meta_id__))
print("Meta id: {}".format(count_words.__meta_id__))

reload_add_numbers = meta.load(add_numbers.__meta_id__)

assert(reload_add_numbers(4,5) == add_numbers(4,5))

print(add_numbers.analytics())

fibonacci = meta.search("fibonacci")

print("fibonacci(3) = {}".format(fibonacci(3)))

for _ in range(0,1000):
    fibonacci(22)

print("duplicates", fibonacci.find_duplicates())

try: fibonacci("s")
except: pass

print("bugs",fibonacci.bugs())

create_dict_from_object = meta.load("http://meta-lang.org/snippet/56ee4528c0cb8f7470fbafa6")

print(create_dict_from_object(meta))

print(meta.overhead_call)
print(meta.overhead_load)

meta.measure_overhead()



import time
import inspect
import dill
import getpass
import numpy as np
#from scipy.stats import ttest_ind
from datetime import datetime
from collections import defaultdict,Counter
from bson.objectid import ObjectId
from bson import Binary
from bson.json_util import loads as bson_loads
from bson.json_util import dumps as bson_dumps
import re
import sys
import os
from .timeout import timeout, timeout_dumb
from collections import defaultdict,Counter
import random
import requests
from . import helpers as util


class WebCache:

    def __init__(self, webpath="http://meta-lang.org"):
        self.path = webpath

    def version_key(self,s):
        return "\t".join([s["name"],s["file"],s["user"]])

    def record_function(self, snippet_data):
        resp = requests.post(self.path + "/record_function", data={"bson":bson_dumps(snippet_data)})
        return resp.json()

    def record_call(self, runtime_info):
        requests.post(self.path + "/record_call", data={"bson":bson_dumps(runtime_info)})

    def record_bug(self, bug_data):
        requests.post(self.path + "/record_bug", data={"bson":bson_dumps(bug_data)})

    def clear_runtime(self, id_):
        requests.post(self.path + "/clear_runtime", data={"id":id_})

    def similar_funcs(self, text, typ=None):
        resp = requests.get(self.path + "/similar_funcs", json={"text":text, "type":typ})
        # resp is a list of snippet json data
        return bson_loads(resp.text)

    def id2func(self, id_):
        resp = requests.get(self.path + "/id_to_func", params={"id":id_})
        # resp is a json snippet
        return bson_loads(resp.text)

    def get_ids(self, ids):
        resp = requests.get(self.path + "/get_ids", json={"ids":ids})
        return bson_loads(resp.text)

    def id2execution(self, id_):
        resp = requests.get(self.path + "/id_to_execution", params={"id":id_})
        # resp is a list of execution json
        return bson_loads(resp.text)

    def id2bugs(self, id_):
        resp = requests.get(self.path + "/id_to_bugs", params={"id":id_})
        return bson_loads(resp.text)

    def update_snippet(self, id_, updates):
        requests.post(self.path + "/update_snippet", data={"id":id_, "bson":bson_dumps(updates)})

    def match_type(self, typ=None, text=None, field="dynamic_type", l=100):
        resp = requests.get(self.path + "/match_type", params={"text":text, "type":typ, "field":field, "limit":l})
        return bson_loads(resp.text)

    def get_all_snippet_ids(self):
        resp = requests.get(self.path + "/all_snippet_ids")
        return bson_loads(resp.text)

    def get_all_docstrings(self):
        resp = requests.get(self.path + "/all_docstrings")
        return bson_loads(resp.text)

class Meta:

    def __init__(self, userid="global", user=None, sandbox=False, debug=False, max_obj_size=20000, backend_url="https://meta-backend.herokuapp.com", overhead=False, globals=globals(), jupyter=False, inspect=inspect):
        self.cache = WebCache(backend_url)
        self.globals = globals
        self.__framework_test__ = sandbox
        self.debug = debug
        self.jupyter = jupyter
        self.inspect = inspect
        self.overhead = overhead
        self.overhead_call = []
        self.overhead_load = []
        self.overhead_create = []
        self.max_obj_size = max_obj_size
        self.overhead_begintime = time.clock()
        self.optimize = bool(os.environ.get("OPT", False))
        self.backoff = util.str_to_bool(os.environ.get("AUTO_PATCH", False))
        if user:
            self.user = user
        else:
            self.user = getpass.getuser()

    def bind(self,id,imports=[]):
        def bind_decorator(func):
            def func_wrapper(*args, **kwargs):
                t_overhead1 = time.clock()
                instrumented = False
                try:
                    t1 = time.clock()
                    v = func(*args, **kwargs)
                    t2 = time.clock()
                except Exception as e:
                    # record bad input
                    bad_input = {"args":Binary(dill.dumps(args)), "snippet_id":id, "time":datetime.now(),
                                 "exception_type":type(e).__name__, "exception":Binary(dill.dumps(e))}
                    self.cache.record_bug(bad_input)
                    raise e
                if random.randint(0,func_wrapper.called) == 0:
                    instrumented = True
                    params = tuple(list(args)+[v])
                    sig = " -> ".join([util.pp_type(x) for x in util.fancy_type(params).__tuple_params__])
                    # args_ = [a if type(a).__name__ != 'generator' else type(a) for a in args]
                    # v_ = v if type(v).__name__ != 'generator' else type(v)
                    #if sys.getsizeof(args) < self.max_obj_size and sys.getsizeof(v) < self.max_obj_size and not type(v).__name__ == 'generator':
                    if not type(v).__name__ == 'generator':
                        runtime = {"snippet_id":id, "call":Binary(dill.dumps((args,v))),"time_running":t2-t1,"file":__file__,
                               "user":self.user,"time":datetime.now(),"type":sig, "imports":imports}
                        if sys.getsizeof(runtime["call"]) < self.max_obj_size:
                            self.cache.record_call(runtime)
                func_wrapper.called += 1
                t_overhead2 = time.clock()
                if self.overhead: self.overhead_call.append({"id":id,"meta_time":t_overhead2-t_overhead1,"normal_time":t2-t1, "inst":instrumented})
                return v
            func_wrapper.called = 0
            return func_wrapper
        return bind_decorator


    def __call__(self, d_str=None, imports=None, parent=None):
        def real_dec(func):
            t_overhead1 = time.clock()
            imports_copy = imports
            src = "".join(self.inspect.getsourcelines(func)[0][1:])
            if imports_copy == None:
                imports_copy = util.list_imports(self.globals,src)
            f_name = func.__name__
            is_undef = util.is_func_undefined(src)
            annote = {k:util.pp_type(v) for k,v in func.__annotations__.items()}
            arg_type = [annote[a] if a in annote else "*" for a in func.__code__.co_varnames]
            ret_type = annote["return"] if "return" in annote else "*"
            t_sig = " -> ".join(arg_type+[ret_type])
            if self.jupyter:
                file_data = "jupyter"
            else:
                file_data = sys.modules[func.__module__].__file__
            # possibly I want to mark when we're recording input fed to a magically shifted function
            snippet_data = {
                "source":src, "imports":imports_copy,
                "name":f_name, "type":t_sig,
                "undefined":is_undef, "doc":d_str,
                "parent":parent,
                "created_at":datetime.now(),
                "file":file_data,
                "user":self.user,
            }
            id_ = self.cache.record_function(snippet_data)
            if is_undef:
                old_func = func
                func = self.search(d_str)
                id_ = func.__meta_id__
            snippet_data["func"] = func
            inner = MetaFunction(id_, self)#, snippet_data)
            t_overhead2 = time.clock()
            if self.overhead: self.overhead_create.append(t_overhead2-t_overhead1)
            return inner
        return real_dec

    def measure_overhead(self):
        overhead_endtime = time.clock()
        len_load = len(self.overhead_load)
        len_create = len(self.overhead_create)
        sum_load = np.sum(self.overhead_load)
        avg_load = np.average(self.overhead_load)
        avg_create = np.average(self.overhead_create)
        sum_create = np.sum(self.overhead_create)
        len_calls = len(self.overhead_call)
        sum_inst_call = np.sum([x["meta_time"] for x in self.overhead_call if x["inst"]])
        sum_inst_baseline = np.sum([x["normal_time"] for x in self.overhead_call if x["inst"]])
        len_inst = len([x for x in self.overhead_call if x["inst"]])
        sum_meta_call = np.sum([x["meta_time"] for x in self.overhead_call])
        sum_normal_call = np.sum([x["normal_time"] for x in self.overhead_call])
        total_unadjusted = overhead_endtime - self.overhead_begintime
        print(total_unadjusted,"total")
        meta_total = total_unadjusted - sum_normal_call
        normal_total = total_unadjusted - sum_load - sum_meta_call - sum_create
        print(meta_total, normal_total)
        print("{} overall overhead".format(meta_total/normal_total))
        print("{} functions loaded and {} created and {} called".format(len_load, len_create, len_calls))
        print("{} overhead per meta call".format((sum_inst_call - sum_inst_baseline)/len_inst))
        print("call overhead as percent {}".format((sum_meta_call - sum_normal_call) / meta_total))
        print("{} overhead per prob meta call".format((sum_meta_call - sum_normal_call)/len_calls))
        print("{} avg load time".format(avg_load))
        print("{} avg create time".format(avg_create))
        data_out = [total_unadjusted, meta_total / normal_total, len_load, len_create, len_calls, ((sum_meta_call - sum_normal_call)/len_calls), avg_load, avg_create, meta_total, normal_total, (sum_meta_call - sum_normal_call), sum_load, sum_create]
        with open("overhead2.tsv", "a") as myfile:
            myfile.write("\t".join([str(x) for x in data_out])+"\n")


    def load(self, id_: str, test=False, instrument=True):
        t_overhead1 = time.clock()
        mf = MetaFunction(id_, self)
        t_overhead2 = time.clock()
        if self.overhead: self.overhead_load.append(t_overhead2-t_overhead1)
        return mf

    def load_multiple(self, ids):
        snippets = self.cache.get_ids(ids)
        loaded_mfuncs = {str(s["_id"]):MetaFunction(str(s["_id"]), self, data=s) for s in snippets}
        # want to maintain order of request
        return [loaded_mfuncs[i] for i in ids]

    def search(self, doc="", type=None, n=1, examples=None):
        if examples and (not type):
            type = util.type_to_string(examples,False)[0]
        possible = self.cache.match_type(type,doc,l=n)
        scores = [(str(p["_id"]), p["doc"], util.query_string_score(doc,p["doc"])) for p in possible]
        sorted_scores = sorted(scores,key=lambda x: x[2],reverse=True)
        get_funcs = self.load_multiple([s[0] for s in sorted_scores]) #[self.load(s[0]) for s in sorted_scores]
        if examples: get_funcs = util.filter_funcs_by_examples(get_funcs, examples)
        if n==1:
            return get_funcs[0]
        else:
            return get_funcs[:n]

    def compare(self, mf1, mf2):
        execution = mf1.analytics()["recent_calls"] + mf2.analytics()["recent_calls"]
        inputs = [x[0] for x in execution]
        total_t1, total_t2 = [], []
        for i in inputs:
            total_t1 += util.average_times(mf1.raw, i, n=100)
            total_t2 += util.average_times(mf2.raw, i, n=100)
        t1_avg, t2_avg = [np.average(x) for x in [total_t1, total_t2]]
        if t1_avg > t2_avg:
            odds = t1_avg / t2_avg
        else:
            odds = t2_avg / t1_avg
        if True:#ttest_ind(total_t1, total_t2)[1] < 0.05:
            return (odds, t1_avg/len(inputs), t2_avg/len(inputs))
        else:
            return None

    def one_step_search(self, b, e, qs):
        start_n = len(b.split("->"))+1
        b_rgx = re.compile("^"+re.escape(b))
        e_rgx = re.compile(re.escape(e)+"$")
        def extract(x): return (x["_id"],x["doc"],x["for_inference"])
        def inflate(x):
            return [(x[0],x[1],[y_.rstrip().lstrip() for y_ in y.split("->")],y) for y in x[2]]
        def pop_begin(lst,n):
            return [(x[0],x[1],x[2][-1],x[3]) for x in lst if len(x[2]) == n]
        def pop_end(lst):
            return [(x[0],x[1],x[2][0],x[3]) for x in lst if len(x[2]) == 2]
        begin = [inflate(y) for y in [extract(x) for x in self.cache.match_type(b_rgx,qs,field="for_inference",l=100)]]
        end = [inflate(y) for y in [extract(x) for x in self.cache.match_type(e_rgx,qs,field="for_inference",l=100)]]
        begin, end = [list(it.chain(*x)) for x in [begin,end]]
        match_sets = []
        possible_ends = pop_end(end)
        for n in range(start_n,5):
            b_ = pop_begin(begin,n)
            match_sets.append([b_,possible_ends])
        results = []
        for m in match_sets:
            for pair in it.product(*m):
                # print(pair[0],pair[1])
                if pair[0][2] == pair[1][2]:
                    comb_str = pair[0][1].lower() + ' and ' + pair[1][1].lower()
                    results.append((pair, util.query_string_score(comb_str,qs), comb_str))
        s_results = sorted(results,key=lambda x: x[1],reverse=True)
        final_results = []
        for r in s_results:
            s1 = str(r[0][0][0])
            s2 = str(r[0][1][0])
            def gen(s1,s2):
                def compile():
                    f1 = MetaFunction(s1,self)
                    f2 = MetaFunction(s2,self)
                    comb = lambda *x: f2(f1(*x))
                    return comb
                return compile
            final_results.append((gen(s1,s2),r[2],r[1]))
        return final_results

    def broad_search(self,text):
        return self.cache.similar_funcs(text)[:25]

    def all_funcs(self):
        for id_ in self.cache.get_all_snippet_ids():
            yield self.load(id_)

class MetaFunction():

    def __init__(self, id_: str, meta_conn, data=None, load_at_all_costs=True):
        id_ = str(id_).split("/")[-1]
        self.__meta_id__ = id_
        self.meta = meta_conn
        self.duplicates = None
        self.__meta_backoff__ = False
        self.__show_search__ = False
        if not data:
            data = self.meta.cache.id2func(id_)
        src, func_name, imports = [data[k] for k in ["source","name","imports"]]
        try:
            new_f = util.load_source(func_name, src, imports)
        except Exception as e:
            if load_at_all_costs:
                new_f = util.any_func
                pass
            else:
                raise e
        if "show_search" in data and data["show_search"] == True:
            self.__show_search__ = True
        self.__meta_doc__ = data["doc"]
        self.__meta_type__ = data["type"]
        self.__meta_source__ = data["source"]
        self.__meta_imports__ = imports
        if "for_inference" in data:
            self.__meta_for_inference__ = data["for_inference"]
        else:
            self.__meta_for_inference__ = None
        if "avg_time" in data:
            self.__meta_avg_time__ = data["avg_time"]
        else:
            avg = self.analytics()['average run time']
            self.__meta_avg_time__ = avg
        self.raw = new_f
        self.sandbox_io = new_f # need to fix sandbox AGAIN
        # self.sandbox_io = iosandbox[func_name] #fake_io(new_f,debug=True)
        # self.instrument_sandbox = self.meta.bind(id_)(self.sandbox_io)
        self.instrument_f = self.meta.bind(id_)(new_f)
        self.__name__ = func_name

    def __repr__(self):
        return self.__meta_doc__ + " (" + self.__name__ + ")"

    def __call__(*args,**kwargs):
        # args[0] = self?
        self = args[0]
        args = args[1:]
        if self.meta.optimize:
            return self.optimize()
        if self.meta.__framework_test__:
            return self.instrument_sandbox(*args,**kwargs)
        if self.meta.backoff or self.__meta_backoff__:
            self.find_duplicates()
            for d in [self]+self.duplicates:
                try:
                    new_ret = d.instrument_f(*args,**kwargs)
                    print("Warning: Auto-patched {} with {}".format(self.__name__, d.__name__))
                    return new_ret
                except:
                    pass
            raise Exception("No backoff function could save this input")
        else:
            ret = self.instrument_f(*args,**kwargs)
            if self.__show_search__ == False:
                self.meta.cache.update_snippet(self.__meta_id__, {"show_search":True})
                self.__show_search__ = True
            return ret

    def get_dynamic_type_sig(self, _expand_ = True, recompute=False):
        # what if no inputs or no calls?
        if self.meta.debug:
            print("get_dynamic_type_sig",self.__meta_doc__,self.__name__)
        execution = [x for x in self.meta.cache.id2execution(self.__meta_id__)]
        # if len(execution) == 0: return []
        if execution:
            parsed = [util.safe_load(x["call"]) for x in execution]
            self.__meta_dynamic_type__ = util.type_to_string(parsed, True)
            self.__meta_for_inference__ = util.type_to_string(parsed, False)
            # update db
            self.meta.cache.update_snippet(self.__meta_id__,
                                           {"dynamic_type":self.__meta_dynamic_type__,
                                            "for_inference":self.__meta_for_inference__})
            if _expand_:
                return self.__meta_for_inference__
            else:
                return self.__meta_dynamic_type__

    def test_as(self, other_meta_func, debug=False, fail_count=float("inf"), io_skip=False):
        #ex1, ex2 = [self.meta.cache.id2execution(x.__meta_id__) for x in [self,other_meta_func]]
        exs = self.meta.cache.id2execution(self.__meta_id__)
        execution = set([x["call"] for x in exs]) #list(ex1)+list(ex2)])
        if len(execution) == 0:
            return 0
        pass_, fail_ = 0.0, 0.0
        try:
            func_1 = timeout_dumb(2)(dill.dumps(self.raw,recurse=True))
            func_2 = timeout_dumb(2)(dill.dumps(other_meta_func.raw,recurse=True))
        except Exception as e:
            print(e,file=sys.stderr)
            return 0
        for i,test in enumerate(execution):
            if fail_ == fail_count:
              return pass_ / (pass_ + fail_)
            # if self.sandbox_io.done_io:
            #     return 0
            if test == None: continue
            call = util.safe_load(test)
            if call == None: continue
            args, ret = call[0], call[1]
            # print(i,args)
            try:
                v1 = func_1(*args)
                v2 = func_2(*args)
                if v1 == None and v2 == None:
                    fail_ += 1.0
                    continue
                if not util.equal_fvals(v1,v2):
                    if debug: print("Failed test {} with args {}: returned {} but expected {}".format(i,args,v1,v2))
                    fail_ += 1.0
                else:
                    if debug: print("Passed test {}, {}".format(i,args))
                    pass_ += 1.0
            except Exception as inst:
                if debug:
                    print("Failed test {} with exception".format(i))
                    print(inst)
                fail_ += 1.0
        return pass_ / (pass_ + fail_)

    def find_duplicates(self):
        if self.duplicates:
            return self.duplicates
        else:
            snippet = self.meta.cache.id2func(self.__meta_id__)
            if "for_inference" in snippet and snippet["for_inference"] and len(snippet["for_inference"]) > 0:
                tp = snippet["for_inference"]
                matches = self.meta.cache.similar_funcs(snippet["doc"], tp)
                to_ret = []
                for m in matches:
                    if str(m["_id"]) == self.__meta_id__: continue
                    #print("test as",str(m["_id"]),str(id_))
                    try:
                        loaded_f = self.meta.load(str(m["_id"]))
                    except:
                        print("failed to load {}".format(m["_id"]))
                        continue
                    #print(loaded_f,file=sys.stdout)
                    score = self.test_as(loaded_f, fail_count=1, debug=False)
                    if score >= 1:
                        to_ret.append(loaded_f)
                self.duplicates = to_ret
                return to_ret
            else:
                self.duplicates = []
                return []

    def optimize(self,k="average run time"):
        dups = self.find_duplicates()
        curr_time = round(self.analytics()[k]*1000,4)
        rank = sorted([(d,round(d.analytics()[k]*1000,4)) for d in dups],key=lambda x: x[1])
        if len(rank)>0 and rank[0][1] < curr_time:
            print("Warning: optimizing {} with {}:\nSee: http://www.meta-lang.org/snippets/{}\nAverage time of {}ms vs. {}ms".format(self.__name__, rank[0][0].__name__, rank[0][0].__meta_id__, rank[0][1], curr_time),file=sys.stderr)
            selected = rank[0][0].raw
        else:
            selected = self
        def opt(*args,**kwargs):
            return selected(*args,**kwargs)
        return opt

    def optimize2(self, n_inputs=10, n_times=10, p_cutoff=0.05):
        dups = self.find_duplicates()
        # can make this more efficient
        all_inputs = [util.safe_load(x)[0] for x in set([x["call"] for x in self.meta.cache.id2execution(self.__meta_id__)])]
        test_inputs = random.sample([x for x in all_inputs if x != None], min(n_inputs,len(all_inputs)))
        self_times = self.time_on_inputs(test_inputs, n_times).flatten()
        self_average = np.average(self_times)
        out = []
        print(dups)
        for d in dups:
            d_times = d.time_on_inputs(test_inputs, n_times).flatten()
            d_average = np.average(d_times)
            print(self_average,d_average)
            if d_average < self_average:
                # print(self_average / d_average)
                # _, pval = ttest_ind(self_times, d_times)
                # if pval < 0.05:
                out.append([d, self_average / d_average])
        return out

    def time_on_inputs(self, inputs, n=10):
        times_per_input = []
        for i in inputs:
            timed_func = util.timing(self.raw)
            times = [timed_func(*i)[0] for _ in range(0,n)]
            times_per_input.append(times)
        return np.array(times_per_input)

    def analytics(self):
        execution = self.meta.cache.id2execution(self.__meta_id__)
        execution = list(execution)
        avg_time = np.average([x["time_running"] for x in execution])
        sorted_exe = sorted(execution,key=lambda x: x["time"],reverse=True)
        all_rets = [util.safe_load(x["call"]) for x in sorted_exe]
        all_types = Counter([util.fancy_type(x) for x in all_rets])
        last_ten = [{"input":x[0], "output": x[1]} for x in all_rets[:10]]
        self.meta.cache.update_snippet(self.__meta_id__,{"avg_time":avg_time})
        return {"average run time":avg_time, "recent_calls":last_ten, "types":all_types}

    def examples(self):
        return self.analytics()["recent_calls"]

    def url(self):
        return "http://www.meta-lang.org/snippets/{}".format(self.__meta_id__)

    def get_type(self):
        def replace_any(x):
            return re.sub(r"\[[^\]]*\]","",x)
        self.get_dynamic_type_sig()
        return list(set([replace_any(x) for x in self.__meta_for_inference__]))
        #return self.__meta_dynamic_type__

    def bugs(self):
        bugs = self.meta.cache.id2bugs(self.__meta_id__)
        out = []
        for b in bugs:
            b["args"] = dill.loads(b["args"])
            b["exception"] = dill.loads(b["exception"])
            out.append(b)
        return out

    def overhead(self,n):
        execution = self.meta.cache.id2execution(self.__meta_id__)
        inputs = [util.safe_load(x["call"])[0] for x in execution]
        time_raw, time_meta = 0.0, 0.0
        c = 0
        for args in inputs:
            for _ in range(0,n):
                sometimes = self.instrument_sandbox
                time_raw += util.average_times(self.sandbox_io,args,n=1)[0]
                if random.randint(0,c) != 0:
                    sometimes = self.sandbox_io
                time_meta += util.average_times(sometimes,args,n=1)[0]
                c += 1
        return time_raw, time_meta

    def backoff(self,pos=True):
        self.__meta_backoff__ = pos

class Test:

    def __init__(self):
        self.meta = Meta()

    def test_type_sigs(self):
        print("Testing dynamic type sig inference".upper())
        for sid in self.meta.cache.get_all_snippet_ids():
            print(sid,self.meta.get_dynamic_type_sig(sid,_expand_=False))
        print("")

    def test_load(self):
        print("Testing function load from id".upper())
        for sid in self.meta.cache.get_all_snippet_ids():
            print(self.meta.load(sid).__name__)
        print("")

    def test_search(self):
        print("Testing function load from docstr".upper())
        for docstr in self.meta.cache.get_all_docstrings():
            print(docstr)
            print(self.meta.search(docstr).__name__)
        print("")

    def test_testgen(self):
        print("Testing generation of test cases".upper())
        for sid in self.meta.cache.get_all_snippet_ids():
            f = self.meta.load(sid,test=True)
            print(f.__name__)
            self.meta.test_as(f,sid)
        print("")

    def test_duplicates(self):
        print("Testing possible duplicate finding")
        for sid in self.meta.cache.get_all_snippet_ids():
            try:
                f = self.meta.load(sid)
            except:
                continue
            dups = f.find_duplicates()
            #print(sid)
            if len(dups) > 0:
                print(sid,self.meta.cache.id2func(sid)["doc"])
                print("Possible duplicates",dups)
                # now do a thing where we actually look at input/output

    def test_typesearch(self,queries=None):
        print("Testing type search".upper())
        if queries == None:
            queries = [
                {"type": "List[int] -> int", "doc": "sum the list"},
                {"type": "int -> str", "doc": "convert int to str"},
                {"type": "int -> int -> int", "doc":"add numbers"}
            ]
        for q in queries:
            print("Searched for '{}' of type '{}'".format(q["doc"],q["type"]))
            funcs = self.meta.type_search(q["type"],q["doc"])
            for tp in funcs:
                print((tp[0].__name__,tp[1],tp[2]))

    def test_overhead(self,n):
        raw_total, meta_total = 0.0, 0.0
        for sid in self.meta.cache.get_all_snippet_ids():
            f = self.meta.load(sid)
            try:
                raw_, meta_ = f.overhead(n)
                raw_total += raw_
                meta_total += meta_
                print(raw_total / meta_total)
            except Exception as e:
                print(e)
                pass


    def test_all(self):
        self.test_type_sigs()
        self.test_load()
        self.test_search()
        #self.test_testgen()
        self.test_typesearch()



from context import Meta

meta = Meta()

merge_two_dicts = meta.load("http://www.meta-lang.org/snippets/56ee471dc0cb8f7470fbb500")

flatten_list = meta.load("http://www.meta-lang.org/snippets/56ee45b9c0cb8f7470fbb185")

unordered = meta.load("http://www.meta-lang.org/snippets/56ee4583c0cb8f7470fbb0c5")

dict1 = {x:i for x,i in enumerate(range(0,200))}
dict2 = {x:i for x,i in enumerate(range(0,200))}

# or pass OPT=True python ...
# merge_two_dicts(dict1,dict2)
# merge_two_dicts.optimize()(dict1,dict2)

print(merge_two_dicts.optimize2())
# print(flatten_list.optimize2())
print(unordered.optimize2())



from distutils.core import setup
setup(
  name = 'metalang',
  packages = ['metalang'], # this must be the same as the name above
  version = '0.51',
  description = 'A domain specific language that enables powerful code sharing',
  author = 'Ethan Fast',
  author_email = 'ejhfast@gmail.com',
  url = 'https://github.com/Ejhfast/meta', # use the URL to the github repo
  download_url = 'https://github.com/Ejhfast/meta/tarball/0.51',
  keywords = ['DSL', 'code sharing', 'metaprogramming'], # arbitrary keywords
  classifiers = [],
  install_requires=[
          'numpy',
          'dill',
          'pymongo',
          'requests',
          'typing',
          'multiprocess'
  ]
)



from context import Meta

meta = Meta(overhead=True)

# we need to put complex stuff in here :(

@meta("make CountVectorizer to encode word data as a vector")
def make_vectorizer(docs, max_ngram):
    from sklearn.feature_extraction.text import CountVectorizer
    cv = CountVectorizer(stop_words='english', ngram_range=(1,max_ngram))
    vecs = cv.fit_transform(docs)
    return {"vector":vecs, "vectorizer":cv}

print(make_vectorizer.__meta_id__)
print(make_vectorizer.__meta_source__)

vecs = make_vectorizer(["hello there","yeah that is right"],1)
vecs = make_vectorizer(["hello there","yeah that is right"],1)
vecs = make_vectorizer(["hello there","yeah that is right"],1)
vecs = make_vectorizer(["hello there","yeah that is right"],1)

print(vecs["vectorizer"])

print(meta.overhead_call)
print(meta.overhead_load)



from .typing import List,Dict,Any,Tuple,Set,Callable,Union
import itertools as it
import dill
import numpy as np
import numpy as np
import time
from .timeout import timeout
import types
import re

# list all imported modules, doesn't work with "as" or "from"
# optionally filter by their presence in src
def list_imports(globals, src=None):
    imps = set()
    for name, val in globals.items():
        if isinstance(val, types.ModuleType):
            name = val.__name__
            if src:
                if re.compile(name+"\.").search(src):
                    imps.add(name)
                else:
                    continue
            imps.add(name)
    return list(imps)

def str_to_bool(s):
    if s == "True": return True
    return False

# see if meta function is currently undefined
def is_func_undefined(source : str) -> bool:
    return source.rstrip().split("\n")[-1].lstrip() == "pass"

# compute structural type representation for a value
def fancy_type(o):
    if type(o) == type([]):
        if len(o) == 0: return List[Any]
        else: return List[Union[tuple([fancy_type(t_) for t_ in o])]]
    elif type(o) == type(()):
        return Tuple[tuple([fancy_type(t_) for t_ in o])]
    elif type(o) == type({}):
        #return Dict[Any,Any]
        if not o:
            return Dict[Any,Any]
        return Dict[Union[tuple([fancy_type(t_) for t_ in o.keys()])],Union[tuple([fancy_type(t_) for t_ in o.values()])]]
    if type(o) == type(set()):
        return Set[Union[tuple([fancy_type(t_) for t_ in o])]]
    if type(o) == type(fancy_type):
        return Callable[...,Any]
    else:
        return type(o)

# remove 'typing.' when printing types
def pp_type(t):
    if type(t).__module__ == 'metalang.typing':
        return str(t).replace("metalang.typing.","")
    elif t == None:
        return "None"
    else:
        return t.__name__

# dumb heuristic for combined query string similarity
def query_string_score(s1,s2):
    s1, s2 = s1.lower(), s2.lower()
    wds1, wds2 = [set(x.split()) for x in [s1, s2]]
    return len(wds1 & wds2) / float(len(wds1) + len(wds2))

# catch exceptions when loading unsupported argument calls
def safe_load(s):
    try:
        o = dill.loads(s)
        return o
    except:
        return None

# return functions that execute as expected on example inputs
def filter_funcs_by_examples(funcs, examples):
    out = []
    for meta_f in funcs:
        is_good = True
        for in_,out_ in examples:
            try:
                test_out = timeout(2)(meta_f.sandbox_io)(*in_)
                if not equal_fvals(test_out,out_):
                    is_good = False
                    break
            except:
                is_good = False
                break
        if is_good:
            out.append(meta_f)
    return out

# normalize certain kinds of return values for equivalence testing
def ret_normalize(r):
    if isinstance(r,dict):
        return ("dict", sorted(dict(r)))
    elif isinstance(r,np.ndarray):
        try:
            return list(r)
        except:
            return r
    else:
        return r

# test whether two meta functions returned the same value
def equal_fvals(x,y):
    # so hacky
    x,y = [ret_normalize(x_) for x_ in [x,y]]
    return str((x)) == str((y))

# So this is kind of crap, but if you pass add_=[], what it does is break up Unions
def expand_type(t,add_=[Any]):
    if type(t) == type(List):
        return add_ + [List[x] for x in expand_type(t.__parameters__[0],add_)]
    elif type(t) == type(Tuple):
        tups = t.__tuple_params__
        expand_tups = [expand_type(t_,add_) for t_ in tups]
        return add_ + [Tuple[c] for c in it.product(*expand_tups)]
    elif type(t) == type(Dict):
        params = t.__parameters__
        expand_params = [expand_type(t_,add_) for t_ in params]
        return add_ + [Dict[c] for c in it.product(*expand_params)]
    elif type(t) == type(Set):
        return add_ + [Set[x] for x in expand_type(t.__parameters__[0],add_)]
    elif type(t) == type(Union):
        tups = t.__union_params__
        expand_tups = [expand_type(t_,add_) for t_ in tups]
        return add_ + [Union[c] for c in it.product(*expand_tups)]+list(it.chain(*expand_tups))
    else:
        return add_ + [t]

# convert a list of types into a string representation of type signature
def type_to_string(params, __expand__=True):
    params = [tuple(list(i[0])+[i[1]]) for i in params if i != None]
    if len(params) == 0: return []
    sig = Union[tuple([fancy_type(t_) for t_ in params])]
    expanded = expand_type(sig) if __expand__ else expand_type(sig,add_=[])
    p_sig = []
    for s in expanded:
        if type(s) == type(Tuple):
            #print(s)
            p_sig.append(" -> ".join([pp_type(x) for x in s.__tuple_params__]))
            #print(p_sig[-1])
    return p_sig

# time a function and prepend time to return value
def timing(f):
    def wrap(*args):
        time1 = time.time()
        ret = f(*args)
        time2 = time.time()
        return ((time2-time1), ret)
    return wrap

# sample many running times for function input
def average_times(f,args,n=10):
    total = []
    for _ in range(0,n):
        t, _ = timing(timeout(1)(f))(*args)
        total.append(t)
    return total

# a meta placeholder function
def any_func(*args, **kwargs): raise Exception("Not real")

# load meta function imports into sandbox
def load_imports(imports,env):
    # LOL, seriously
    exec("from typing import List,Dict,Any,Tuple,Set,Callable,Union",env)
    for i in imports:
        if type(i) == type(""):
            exec("import {}".format(i),env)
        elif type(i) == type({}):
            exec("from {} import {}".format(i["from"],i["import"]),env)

# load function source and sandbox imports
def load_source(func_name, source, imports):
    sandbox = {}
    load_imports(imports, sandbox)
    exec(source, sandbox)
    return sandbox[func_name]



from context import Meta

meta = Meta()

@meta("convert bool to int")
def bool_to_int(b):
    if b: return 1
    return 0

test = ["True" for _ in range(0,1000)]

# this should not take forever
for t in test:
    bool_to_int(t)
